using System;
using System.IO;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Net;

namespace Mice {

	public enum Note : byte {
		None,
		Move,
	}

	public class Player {
		Point location;
		public Point Location { get { return location; } }
		public Vector2 Position { get { return new Vector2(location.X, location.Y); } }
		public void Apply(Stream stream) {
			stream.Position = 0;

			using (BinaryReader reader = new BinaryReader(stream)) {
				Apply(reader);
			}
		}
		public void Apply(BinaryReader reader) {
			switch ((Note)reader.ReadByte()) {
			case Note.Move: {
					location.X = reader.ReadInt32();
					location.Y = reader.ReadInt32();
				}
				break;
			}
		}
	}

	public class CommunityView : DrawableGameComponent {
		SpriteFont font;
		SpriteBatch spriteBatch;
		readonly Community community;
		Texture2D texture;

		public CommunityView(Game game, Community community)
			: base(game) {
			this.community = community;
		}
		protected override void LoadContent() {
			spriteBatch = new SpriteBatch(GraphicsDevice);
			font = Game.Content.Load<SpriteFont>("font/default");
			texture = Game.Content.Load<Texture2D>("texture/mouse");
			base.LoadContent();
		}

		string actionText = string.Empty;
		public string ActionText {
			get { return actionText; }
			set { actionText = value; }
		}

		public override void Draw(GameTime gameTime) {
			NetworkSession session = community.Session;
			spriteBatch.Begin();
			MouseState state = Mouse.GetState();

			spriteBatch.DrawString(font, ActionText, new Vector2(0x20, 0x00), Color.Orange);
			spriteBatch.DrawString(font, community.State, new Vector2(0x20, 0x20), Color.Red);
			if (session != null) {
				spriteBatch.DrawString(font, string.Format("Host:{0}", session.IsHost), new Vector2(0x20, 0x40), Color.Red);

				int i = 0;
				foreach (NetworkGamer gamer in session.AllGamers) {
					spriteBatch.DrawString(font, string.Format("Gamer:{0} RoundTripTime:{1}",
						gamer, gamer.RoundtripTime),
						new Vector2(0x20, 0x60 + 0x20 * i), Color.White);

					Player player = gamer.Tag as Player;
					if (player != null) {
						spriteBatch.Draw(texture, player.Position
							- new Vector2(texture.Width * 0.5f, texture.Height * 0.5f), Color.White);
						Vector2 position = player.Position
							+ font.MeasureString(gamer.Gamertag) * new Vector2(-0.5f, -1.0f)
							- new Vector2(0, texture.Height * 0.5f);
						spriteBatch.DrawString(font, gamer.Gamertag, position, Color.White);
					}
					i++;
				}

			} else {
				spriteBatch.DrawString(font, string.Format("No Session"), new Vector2(0x20, 0x40), Color.Red);
			}
			spriteBatch.End();

			base.Draw(gameTime);
		}
	}

	public class Community : DrawableGameComponent {
		public Community(Game game) : base(game) { }
		volatile
		NetworkSession session;
		volatile
		IAsyncResult result;

		string state = string.Empty;
		/// <summary>
		/// ~ςĂM1t[Ŏoő吔B
		/// </summary>
		const int MaximumCatchupFrames = 10;

		bool requestOnline = true;

		public string State {
			get { return state; }
			set { state = value; }
		}

		public void Connect() {
			requestOnline = true;
		}
		public void Disconnect() {
			requestOnline = false;
		}

		void BeginFind() {
			if (Gamer.SignedInGamers.Count > 0 && result == null) {
				NetworkSessionProperties properties = new NetworkSessionProperties();
				result = NetworkSession.BeginFind(NetworkSessionType.SystemLink, 4, properties, new AsyncCallback(EndFind), null);
				state = "Session Finding";
			} else {
				result = null;
			}
		}

		public NetworkSession Session {
			set {
				if (session != null) {
					session.GamerJoined -= GamerJoined;
				}
				session = value;
				if (session != null) {
					session.GamerJoined += GamerJoined;
				}
			}
			get { return session; }
		}

		void EndFind(IAsyncResult result) {
			try {
				using (AvailableNetworkSessionCollection sessions = NetworkSession.EndFind(result)) {
					this.result = null;
					if (sessions.Count > 0) {
						Session = NetworkSession.Join(sessions[0]);
						state = "Session Join";
					} else
						BeginCreate();
				}
			} catch (Exception ex) {
				state = ex.Message;
				this.result = null;
			}
		}
		void BeginCreate() {
			if (Gamer.SignedInGamers.Count > 0 && result == null) {
				result = NetworkSession.BeginCreate(NetworkSessionType.SystemLink, 2, 4, 1, new NetworkSessionProperties(), new AsyncCallback(EndCreate), null);
				state = "Session Creating";
			}
		}
		void EndCreate(IAsyncResult result) {
			Session = NetworkSession.EndCreate(result);
			session.AllowJoinInProgress = true;
			state = "Session Created";
			this.result = null;
		}
		int wait;
		public override void Update(GameTime gameTime) {
			if (session == null) {
				if (++wait >= 60 * 4) {
					if (requestOnline) {
						BeginFind();
						wait = 0;
					}
				}
			} else {
				session.Update();

				MouseState state = Mouse.GetState();
				MemoryStream stream = new MemoryStream();
				BinaryWriter w = new BinaryWriter(stream);
				w.Write((byte)Note.Move);
				w.Write(state.X);
				w.Write(state.Y);


				foreach (LocalNetworkGamer gamer in session.LocalGamers) {
					PacketWriter writer = new PacketWriter();
					writer.Write(stream.GetBuffer(), 0, (int)stream.Length);
					gamer.SendData(writer, SendDataOptions.InOrder);
				}

				PacketReader reader = new PacketReader();
				foreach (LocalNetworkGamer gamer in session.LocalGamers) {
					NetworkGamer sender;
					for (int i = 0; i < MaximumCatchupFrames; i++) {
						if (gamer.IsDataAvailable) {
							gamer.ReceiveData(reader, out sender);
							Player player = sender.Tag as Player;
							if (player != null)
								player.Apply(reader);
						}
					}
				}

				if (++wait >= 60 * 4) {
					if (!requestOnline) {
						session.Dispose();
						session = null;
						wait = 0;
					}
				}
			}

			base.Update(gameTime);
		}

		void GamerJoined(object sender, GamerJoinedEventArgs e) {
			e.Gamer.Tag = new Player();
		}
	}

	public class CommunityGame : Microsoft.Xna.Framework.Game {
		readonly GraphicsDeviceManager graphics;
		readonly Community community;
		readonly CommunityView communityView;

		public CommunityGame() {
			this.TargetElapsedTime = TimeSpan.FromMilliseconds(100);
			graphics = new GraphicsDeviceManager(this);
			Content.RootDirectory = "Content";
			Components.Add(new GamerServicesComponent(this));
			Components.Add(community = new Community(this));
			Components.Add(communityView = new CommunityView(this, community));
			communityView.ActionText = "Last Action: None / [F1] Disconnect, [F2] Connect";
		}

		protected override void Update(GameTime gameTime) {
			KeyboardState state = Keyboard.GetState();
			if (state.IsKeyDown(Keys.F1)) {
				communityView.ActionText = "Last Action: Disconnect";
				community.Disconnect();
			}
			if (state.IsKeyDown(Keys.F2)) {
				communityView.ActionText = "Last Action: Connect";
				community.Connect();
			}
			IsMouseVisible = community.Session == null;

			base.Update(gameTime);
		}
		protected override void Draw(GameTime gameTime) {
			GraphicsDevice.Clear(ClearOptions.Target | ClearOptions.DepthBuffer, Color.Green, 1.0f, 0);
			base.Draw(gameTime);
		}

	}
}
